@@ -48,6 +48,7 @@
 #include "util/ralloc.h"
 #include "i915/iris_bufmgr.h"
 #include "iris_batch.h"
+#include "iris_bufmgr.h"
 #include "iris_context.h"
 #include "iris_resource.h"
 #include "iris_screen.h"
@@ -69,6 +70,8 @@ enum modifier_priority {
    MODIFIER_PRIORITY_4_DG2_RC_CCS_CC,
    MODIFIER_PRIORITY_4_MTL_RC_CCS,
    MODIFIER_PRIORITY_4_MTL_RC_CCS_CC,
+   MODIFIER_PRIORITY_4_LNL_CCS,
+   MODIFIER_PRIORITY_4_BMG_CCS,
 };
 
 static const uint64_t priority_to_modifier[] = {
@@ -84,6 +87,8 @@ static const uint64_t priority_to_modifier[] = {
    [MODIFIER_PRIORITY_4_DG2_RC_CCS_CC] = I915_FORMAT_MOD_4_TILED_DG2_RC_CCS_CC,
    [MODIFIER_PRIORITY_4_MTL_RC_CCS] = I915_FORMAT_MOD_4_TILED_MTL_RC_CCS,
    [MODIFIER_PRIORITY_4_MTL_RC_CCS_CC] = I915_FORMAT_MOD_4_TILED_MTL_RC_CCS_CC,
+   [MODIFIER_PRIORITY_4_LNL_CCS] = I915_FORMAT_MOD_4_TILED_LNL_CCS,
+   [MODIFIER_PRIORITY_4_BMG_CCS] = I915_FORMAT_MOD_4_TILED_BMG_CCS,
 };
 
 static bool
@@ -128,11 +133,20 @@ modifier_is_supported(const struct intel_device_info *devinfo,
       if (!intel_device_info_is_mtl_or_arl(devinfo))
          return false;
       break;
+   case I915_FORMAT_MOD_4_TILED_LNL_CCS:
+      if (devinfo->platform != INTEL_PLATFORM_LNL)
+         return false;
+      break;
+   case I915_FORMAT_MOD_4_TILED_BMG_CCS:
+      if (devinfo->platform != INTEL_PLATFORM_BMG)
+         return false;
+      break;
    case DRM_FORMAT_MOD_INVALID:
    default:
       return false;
    }
 
+   bool no_fc = INTEL_DEBUG(DEBUG_NO_FAST_CLEAR);
    bool no_ccs = INTEL_DEBUG(DEBUG_NO_CCS) || (bind & PIPE_BIND_CONST_BW);
 
    /* Check remaining requirements. */
@@ -156,16 +170,22 @@ modifier_is_supported(const struct intel_device_info *devinfo,
          return false;
       }
       break;
-   case I915_FORMAT_MOD_4_TILED_MTL_RC_CCS:
    case I915_FORMAT_MOD_4_TILED_MTL_RC_CCS_CC:
    case I915_FORMAT_MOD_4_TILED_DG2_RC_CCS_CC:
-   case I915_FORMAT_MOD_4_TILED_DG2_RC_CCS:
    case I915_FORMAT_MOD_Y_TILED_GEN12_RC_CCS_CC:
+      if (no_fc)
+         return false;
+      FALLTHROUGH;
+   case I915_FORMAT_MOD_4_TILED_LNL_CCS:
+   case I915_FORMAT_MOD_4_TILED_BMG_CCS:
+   case I915_FORMAT_MOD_4_TILED_MTL_RC_CCS:
+   case I915_FORMAT_MOD_4_TILED_DG2_RC_CCS:
    case I915_FORMAT_MOD_Y_TILED_GEN12_RC_CCS:
    case I915_FORMAT_MOD_Y_TILED_CCS: {
       if (no_ccs)
          return false;
 
+      /* TODO: Do we still face these restrictions on Xe2+? */
       enum isl_format rt_format =
          iris_format_for_usage(devinfo, pfmt,
                                ISL_SURF_USAGE_RENDER_TARGET_BIT).fmt;
@@ -196,6 +216,12 @@ select_best_modifier(const struct intel_device_info *devinfo,
          continue;
 
       switch (modifiers[i]) {
+      case I915_FORMAT_MOD_4_TILED_BMG_CCS:
+         prio = MAX2(prio, MODIFIER_PRIORITY_4_BMG_CCS);
+         break;
+      case I915_FORMAT_MOD_4_TILED_LNL_CCS:
+         prio = MAX2(prio, MODIFIER_PRIORITY_4_LNL_CCS);
+         break;
       case I915_FORMAT_MOD_4_TILED_MTL_RC_CCS_CC:
          prio = MAX2(prio, MODIFIER_PRIORITY_4_MTL_RC_CCS_CC);
          break;
@@ -272,6 +298,8 @@ iris_query_dmabuf_modifiers(struct pipe_screen *pscreen,
       I915_FORMAT_MOD_4_TILED_MTL_RC_CCS,
       I915_FORMAT_MOD_4_TILED_MTL_RC_CCS_CC,
       I915_FORMAT_MOD_4_TILED_MTL_MC_CCS,
+      I915_FORMAT_MOD_4_TILED_LNL_CCS,
+      I915_FORMAT_MOD_4_TILED_BMG_CCS,
       I915_FORMAT_MOD_Y_TILED,
       I915_FORMAT_MOD_Y_TILED_CCS,
       I915_FORMAT_MOD_Y_TILED_GEN12_RC_CCS,
@@ -336,6 +364,8 @@ iris_get_dmabuf_modifier_planes(struct pipe_screen *pscreen, uint64_t modifier,
    case I915_FORMAT_MOD_Y_TILED_GEN12_RC_CCS:
    case I915_FORMAT_MOD_Y_TILED_CCS:
       return 2 * planes;
+   case I915_FORMAT_MOD_4_TILED_LNL_CCS:
+   case I915_FORMAT_MOD_4_TILED_BMG_CCS:
    case I915_FORMAT_MOD_4_TILED_DG2_RC_CCS:
    case I915_FORMAT_MOD_4_TILED_DG2_MC_CCS:
    default:
@@ -381,8 +411,9 @@ iris_memobj_create_from_handle(struct pipe_screen *pscreen,
 
    assert(whandle->type == WINSYS_HANDLE_TYPE_FD);
    assert(whandle->modifier == DRM_FORMAT_MOD_INVALID);
+   /* There is no information if memobj is protected or not */
    struct iris_bo *bo = iris_bo_import_dmabuf(screen->bufmgr, whandle->handle,
-                                              DRM_FORMAT_MOD_INVALID);
+                                              DRM_FORMAT_MOD_INVALID, 0);
    if (!bo) {
       free(memobj);
       return NULL;
@@ -461,7 +492,7 @@ iris_resource_disable_aux(struct iris_resource *res)
    res->aux.state = NULL;
 }
 
-static unsigned
+static enum bo_alloc_flags
 iris_resource_alloc_flags(const struct iris_screen *screen,
                           const struct pipe_resource *templ,
                           struct iris_resource *res)
@@ -473,7 +504,7 @@ iris_resource_alloc_flags(const struct iris_screen *screen,
 
    switch (templ->usage) {
    case PIPE_USAGE_STAGING:
-      flags |= BO_ALLOC_SMEM | BO_ALLOC_COHERENT;
+      flags |= BO_ALLOC_SMEM | BO_ALLOC_CACHED_COHERENT;
       break;
    case PIPE_USAGE_STREAM:
       flags |= BO_ALLOC_SMEM;
@@ -488,9 +519,12 @@ iris_resource_alloc_flags(const struct iris_screen *screen,
    if (templ->bind & PIPE_BIND_SCANOUT)
       flags |= BO_ALLOC_SCANOUT;
 
+   if (templ->flags & PIPE_RESOURCE_FLAG_FRONTEND_VM)
+      flags |= BO_ALLOC_NO_SUBALLOC | BO_ALLOC_NO_VMA;
+
    if (templ->flags & (PIPE_RESOURCE_FLAG_MAP_COHERENT |
                        PIPE_RESOURCE_FLAG_MAP_PERSISTENT))
-      flags |= BO_ALLOC_SMEM | BO_ALLOC_COHERENT;
+      flags |= BO_ALLOC_SMEM | BO_ALLOC_CACHED_COHERENT;
 
    if (screen->devinfo->verx10 >= 125 && screen->devinfo->has_local_mem &&
        isl_aux_usage_has_ccs(res->aux.usage)) {
@@ -532,6 +566,9 @@ iris_resource_destroy(struct pipe_screen *screen,
    if (p_res->target == PIPE_BUFFER)
       util_range_destroy(&res->valid_buffer_range);
 
+   if (p_res->flags & PIPE_RESOURCE_FLAG_FRONTEND_VM)
+      assert(res->bo->address == 0);
+
    iris_resource_disable_aux(res);
 
    threaded_resource_deinit(p_res);
@@ -545,7 +582,7 @@ static struct iris_resource *
 iris_alloc_resource(struct pipe_screen *pscreen,
                     const struct pipe_resource *templ)
 {
-   struct iris_resource *res = calloc(1, sizeof(struct iris_resource));
+   struct iris_resource *res = CALLOC_STRUCT(iris_resource);
    if (!res)
       return NULL;
 
@@ -673,6 +710,26 @@ want_ccs_e_for_format(const struct intel_device_info *devinfo,
    return true;
 }
 
+static bool
+want_hiz_wt_for_res(const struct intel_device_info *devinfo,
+                    const struct iris_resource *res)
+{
+   /* Gen12 only supports single-sampled while Gen20+ supports
+    * multi-sampled images.
+    */
+   if (devinfo->ver < 20 && res->surf.samples > 1)
+      return false;
+
+   if (!(res->surf.usage & ISL_SURF_USAGE_TEXTURE_BIT))
+      return false;
+
+   /* If this resource has the maximum number of samples supported by
+    * running platform and will be used as a texture, put the HiZ surface
+    * in write-through mode so that we can sample from it.
+    */
+   return true;
+}
+
 static enum isl_surf_dim
 target_to_isl_surf_dim(enum pipe_texture_target target)
 {
@@ -861,12 +918,7 @@ iris_resource_configure_aux(struct iris_screen *screen,
       assert(!has_mcs);
       if (!has_ccs) {
          res->aux.usage = ISL_AUX_USAGE_HIZ;
-      } else if (res->surf.samples == 1 &&
-                 (res->surf.usage & ISL_SURF_USAGE_TEXTURE_BIT)) {
-         /* If this resource is single-sampled and will be used as a texture,
-          * put the HiZ surface in write-through mode so that we can sample
-          * from it.
-          */
+      } else if (want_hiz_wt_for_res(devinfo, res)) {
          res->aux.usage = ISL_AUX_USAGE_HIZ_CCS_WT;
       } else {
          res->aux.usage = ISL_AUX_USAGE_HIZ_CCS;
@@ -1011,7 +1063,7 @@ static bool
 iris_resource_image_is_pat_compressible(const struct iris_screen *screen,
                                         const struct pipe_resource *templ,
                                         struct iris_resource *res,
-                                        unsigned flags)
+                                        enum bo_alloc_flags flags)
 {
    assert(templ->target != PIPE_BUFFER);
 
@@ -1022,7 +1074,7 @@ iris_resource_image_is_pat_compressible(const struct iris_screen *screen,
       return false;
 
    if (flags & (BO_ALLOC_PROTECTED |
-                BO_ALLOC_COHERENT |
+                BO_ALLOC_CACHED_COHERENT |
                 BO_ALLOC_CPU_VISIBLE))
       return false;
 
@@ -1030,16 +1082,7 @@ iris_resource_image_is_pat_compressible(const struct iris_screen *screen,
    if ((iris_bufmgr_vram_size(bufmgr) > 0) && (flags & BO_ALLOC_SMEM))
       return false;
 
-   /* We don't have modifiers with compression enabled on Xe2 so far. */
-   if (res->mod_info) {
-      assert(!isl_drm_modifier_has_aux(res->mod_info->modifier));
-      return false;
-   }
-
-   /* TODO: Enable compression on depth surfaces.
-    * https://gitlab.freedesktop.org/mesa/mesa/-/issues/11361
-    */
-   if (isl_surf_usage_is_depth(res->surf.usage))
+   if (res->mod_info && !isl_drm_modifier_has_aux(res->mod_info->modifier))
       return false;
 
    /* Bspec 58797 (r58646):
@@ -1116,7 +1159,7 @@ iris_resource_create_for_image(struct pipe_screen *pscreen,
    const char *name = "miptree";
    enum iris_memory_zone memzone = IRIS_MEMZONE_OTHER;
 
-   unsigned flags = iris_resource_alloc_flags(screen, templ, res);
+   enum bo_alloc_flags flags = iris_resource_alloc_flags(screen, templ, res);
 
    if (iris_resource_image_is_pat_compressible(screen, templ, res, flags))
       flags |= BO_ALLOC_COMPRESSED;
@@ -1146,14 +1189,14 @@ iris_resource_create_for_image(struct pipe_screen *pscreen,
                 res->surf.size_B / INTEL_AUX_MAP_MAIN_SIZE_SCALEDOWN;
    }
 
-   /* Allocate space for the indirect clear color.
-    *
-    * Also add some padding to make sure the fast clear color state buffer
-    * starts at a 4K alignment. We believe that 256B might be enough, but due
-    * to lack of testing we will leave this as 4K for now.
-    */
+   /* Allocate space for the indirect clear color. */
    if (iris_get_aux_clear_color_state_size(screen, res) > 0) {
-      res->aux.clear_color_offset = align64(bo_size, 4096);
+      /* Kernel expects a 4k alignment, otherwise the display rejects the
+       * surface.
+       */
+      const uint64_t clear_color_alignment =
+         (res->mod_info && res->mod_info->supports_clear_color) ? 4096 : 64;
+      res->aux.clear_color_offset = align64(bo_size, clear_color_alignment);
       bo_size = res->aux.clear_color_offset +
                 iris_get_aux_clear_color_state_size(screen, res);
    }
@@ -1220,12 +1263,6 @@ iris_resource_from_user_memory(struct pipe_screen *pscreen,
                                const struct pipe_resource *templ,
                                void *user_memory)
 {
-   struct iris_screen *screen = (struct iris_screen *)pscreen;
-   struct iris_bufmgr *bufmgr = screen->bufmgr;
-   struct iris_resource *res = iris_alloc_resource(pscreen, templ);
-   if (!res)
-      return NULL;
-
    if (templ->target != PIPE_BUFFER &&
        templ->target != PIPE_TEXTURE_1D &&
        templ->target != PIPE_TEXTURE_2D)
@@ -1234,6 +1271,13 @@ iris_resource_from_user_memory(struct pipe_screen *pscreen,
    if (templ->array_size > 1)
       return NULL;
 
+   struct iris_screen *screen = (struct iris_screen *)pscreen;
+   struct iris_bufmgr *bufmgr = screen->bufmgr;
+   struct iris_resource *res = iris_alloc_resource(pscreen, templ);
+   unsigned flags = 0;
+   if (!res)
+      return NULL;
+
    size_t res_size = templ->width0;
    if (templ->target != PIPE_BUFFER) {
       const uint32_t row_pitch_B =
@@ -1262,10 +1306,13 @@ iris_resource_from_user_memory(struct pipe_screen *pscreen,
    size_t mem_size = offset + res_size;
    mem_size = ALIGN_NPOT(mem_size, page_size);
 
+   if (templ->flags & PIPE_RESOURCE_FLAG_FRONTEND_VM)
+      flags |= BO_ALLOC_NO_VMA;
+
    res->internal_format = templ->format;
    res->base.is_user_ptr = true;
    res->bo = iris_bo_create_userptr(bufmgr, "user", mem_start, mem_size,
-                                    IRIS_MEMZONE_OTHER);
+                                    flags, IRIS_MEMZONE_OTHER);
    res->offset = offset;
    if (!res->bo) {
       iris_resource_destroy(pscreen, &res->base.b);
@@ -1315,6 +1362,7 @@ iris_resource_from_handle(struct pipe_screen *pscreen,
    struct iris_screen *screen = (struct iris_screen *)pscreen;
    const struct intel_device_info *devinfo = screen->devinfo;
    struct iris_bufmgr *bufmgr = screen->bufmgr;
+   unsigned flags = 0;
 
    /* The gallium dri layer creates a pipe resource for each plane specified
     * by the format and modifier. Once all planes are present, we will merge
@@ -1325,14 +1373,17 @@ iris_resource_from_handle(struct pipe_screen *pscreen,
    if (!res)
       return NULL;
 
+   if (templ->bind & PIPE_BIND_PROTECTED)
+      flags |= BO_ALLOC_PROTECTED;
+
    switch (whandle->type) {
    case WINSYS_HANDLE_TYPE_FD:
       res->bo = iris_bo_import_dmabuf(bufmgr, whandle->handle,
-                                      whandle->modifier);
+                                      whandle->modifier, flags);
       break;
    case WINSYS_HANDLE_TYPE_SHARED:
       res->bo = iris_bo_gem_create_from_name(bufmgr, "winsys image",
-                                             whandle->handle);
+                                             whandle->handle, flags);
       break;
    default:
       unreachable("invalid winsys handle type");
@@ -1433,18 +1484,13 @@ iris_resource_from_handle(struct pipe_screen *pscreen,
                   goto fail;
             }
 
-            /* Add on a clear color BO if needed.
-             *
-             * Also add some padding to make sure the fast clear color state
-             * buffer starts at a 4K alignment to avoid some unknown issues.
-             * See the matching comment in iris_resource_create_for_image().
-             */
+            /* Add on a clear color BO if needed. */
             if (!main_res->mod_info->supports_clear_color &&
                 iris_get_aux_clear_color_state_size(screen, main_res) > 0) {
                main_res->aux.clear_color_bo =
                   iris_bo_alloc(screen->bufmgr, "clear color buffer",
                                 screen->isl_dev.ss.clear_color_state_size,
-                                4096, IRIS_MEMZONE_OTHER, BO_ALLOC_ZEROED);
+                                64, IRIS_MEMZONE_OTHER, BO_ALLOC_ZEROED);
                if (!main_res->aux.clear_color_bo)
                   goto fail;
             }
@@ -1630,18 +1676,22 @@ iris_flush_resource(struct pipe_context *ctx, struct pipe_resource *resource)
    struct iris_context *ice = (struct iris_context *)ctx;
    struct iris_resource *res = (void *) resource;
    const struct isl_drm_modifier_info *mod = res->mod_info;
-   bool newly_external = false;
-
    /* flush_resource() may be used to prepare an image for sharing externally
     * with other clients (e.g. via eglCreateImage).  To account for this, we
     * make sure to eliminate suballocation and any compression that a consumer
     * wouldn't know how to handle.
+    *
+    * On Xe2+ platforms, when an image wasn't created with a modifier that
+    * supports compression, we need to resolve by copying the image to an
+    * uncompressed bo.
     */
-   if (!iris_bo_is_real(res->bo)) {
+   bool need_pat_resolve = iris_heap_is_compressed(res->bo->real.heap) &&
+                           !(res->base.b.bind & PIPE_BIND_SHARED);
+   bool need_reallocate = !iris_bo_is_real(res->bo) || need_pat_resolve;
+   if (need_reallocate) {
       assert(!(res->base.b.bind & PIPE_BIND_SHARED));
       iris_reallocate_resource_inplace(ice, res, PIPE_BIND_SHARED);
       assert(res->base.b.bind & PIPE_BIND_SHARED);
-      newly_external = true;
    }
 
    iris_resource_prepare_access(ice, res,
@@ -1652,7 +1702,7 @@ iris_flush_resource(struct pipe_context *ctx, struct pipe_resource *resource)
 
    bool disable_aux = !res->mod_info && res->aux.usage != ISL_AUX_USAGE_NONE;
 
-   if (newly_external || disable_aux) {
+   if (need_reallocate || disable_aux) {
       iris_foreach_batch(ice, batch) {
          if (iris_batch_references(batch, res->bo))
             iris_batch_flush(batch);
@@ -1879,6 +1929,9 @@ iris_resource_get_handle(struct pipe_screen *pscreen,
    }
 #endif
 
+   /* TODO: TILE64 modifier support in the KMD */
+   assert(res->surf.tiling != ISL_TILING_64);
+
    switch (whandle->type) {
    case WINSYS_HANDLE_TYPE_SHARED:
       iris_gem_set_tiling(bo, &res->surf);
@@ -1957,7 +2010,9 @@ iris_invalidate_buffer(struct iris_context *ice, struct iris_resource *res)
 {
    struct iris_screen *screen = (void *) ice->ctx.screen;
 
-   if (res->base.b.target != PIPE_BUFFER)
+   if (res->base.b.target != PIPE_BUFFER ||
+       res->base.b.flags & PIPE_RESOURCE_FLAG_FIXED_ADDRESS ||
+       res->base.b.flags & PIPE_RESOURCE_FLAG_FRONTEND_VM)
       return false;
 
    /* If it's already invalidated, don't bother doing anything.
@@ -1987,7 +2042,7 @@ iris_invalidate_buffer(struct iris_context *ice, struct iris_resource *res)
       return false;
 
    struct iris_bo *old_bo = res->bo;
-   unsigned flags = old_bo->real.protected ? BO_ALLOC_PROTECTED : BO_ALLOC_PLAIN;
+   enum bo_alloc_flags flags = old_bo->real.protected ? BO_ALLOC_PROTECTED : BO_ALLOC_PLAIN;
    struct iris_bo *new_bo =
       iris_bo_alloc(screen->bufmgr, res->bo->name, res->base.b.width0,
                     iris_buffer_alignment(res->base.b.width0),
@@ -2165,130 +2220,6 @@ get_image_offset_el(const struct isl_surf *surf, unsigned level, unsigned z,
    assert(z0_el == 0 && a0_el == 0);
 }
 
-/**
- * Get pointer offset into stencil buffer.
- *
- * The stencil buffer is W tiled. Since the GTT is incapable of W fencing, we
- * must decode the tile's layout in software.
- *
- * See
- *   - PRM, 2011 Sandy Bridge, Volume 1, Part 2, Section 4.5.2.1 W-Major Tile
- *     Format.
- *   - PRM, 2011 Sandy Bridge, Volume 1, Part 2, Section 4.5.3 Tiling Algorithm
- *
- * Even though the returned offset is always positive, the return type is
- * signed due to
- *    commit e8b1c6d6f55f5be3bef25084fdd8b6127517e137
- *    mesa: Fix return type of  _mesa_get_format_bytes() (#37351)
- */
-static intptr_t
-s8_offset(uint32_t stride, uint32_t x, uint32_t y)
-{
-   uint32_t tile_size = 4096;
-   uint32_t tile_width = 64;
-   uint32_t tile_height = 64;
-   uint32_t row_size = 64 * stride / 2; /* Two rows are interleaved. */
-
-   uint32_t tile_x = x / tile_width;
-   uint32_t tile_y = y / tile_height;
-
-   /* The byte's address relative to the tile's base addres. */
-   uint32_t byte_x = x % tile_width;
-   uint32_t byte_y = y % tile_height;
-
-   uintptr_t u = tile_y * row_size
-               + tile_x * tile_size
-               + 512 * (byte_x / 8)
-               +  64 * (byte_y / 8)
-               +  32 * ((byte_y / 4) % 2)
-               +  16 * ((byte_x / 4) % 2)
-               +   8 * ((byte_y / 2) % 2)
-               +   4 * ((byte_x / 2) % 2)
-               +   2 * (byte_y % 2)
-               +   1 * (byte_x % 2);
-
-   return u;
-}
-
-static void
-iris_unmap_s8(struct iris_transfer *map)
-{
-   struct pipe_transfer *xfer = &map->base.b;
-   const struct pipe_box *box = &xfer->box;
-   struct iris_resource *res = (struct iris_resource *) xfer->resource;
-   struct isl_surf *surf = &res->surf;
-
-   if (xfer->usage & PIPE_MAP_WRITE) {
-      uint8_t *untiled_s8_map = map->ptr;
-      uint8_t *tiled_s8_map = res->offset +
-         iris_bo_map(map->dbg, res->bo, (xfer->usage | MAP_RAW) & MAP_FLAGS);
-
-      for (int s = 0; s < box->depth; s++) {
-         unsigned x0_el, y0_el;
-         get_image_offset_el(surf, xfer->level, box->z + s, &x0_el, &y0_el);
-
-         for (uint32_t y = 0; y < box->height; y++) {
-            for (uint32_t x = 0; x < box->width; x++) {
-               ptrdiff_t offset = s8_offset(surf->row_pitch_B,
-                                            x0_el + box->x + x,
-                                            y0_el + box->y + y);
-               tiled_s8_map[offset] =
-                  untiled_s8_map[s * xfer->layer_stride + y * xfer->stride + x];
-            }
-         }
-      }
-   }
-
-   free(map->buffer);
-}
-
-static void
-iris_map_s8(struct iris_transfer *map)
-{
-   struct pipe_transfer *xfer = &map->base.b;
-   const struct pipe_box *box = &xfer->box;
-   struct iris_resource *res = (struct iris_resource *) xfer->resource;
-   struct isl_surf *surf = &res->surf;
-
-   xfer->stride = surf->row_pitch_B;
-   xfer->layer_stride = xfer->stride * box->height;
-
-   /* The tiling and detiling functions require that the linear buffer has
-    * a 16-byte alignment (that is, its `x0` is 16-byte aligned).  Here we
-    * over-allocate the linear buffer to get the proper alignment.
-    */
-   map->buffer = map->ptr = malloc(xfer->layer_stride * box->depth);
-   assert(map->buffer);
-
-   /* One of either READ_BIT or WRITE_BIT or both is set.  READ_BIT implies no
-    * INVALIDATE_RANGE_BIT.  WRITE_BIT needs the original values read in unless
-    * invalidate is set, since we'll be writing the whole rectangle from our
-    * temporary buffer back out.
-    */
-   if (xfer->usage & PIPE_MAP_READ) {
-      uint8_t *untiled_s8_map = map->ptr;
-      uint8_t *tiled_s8_map = res->offset +
-         iris_bo_map(map->dbg, res->bo, (xfer->usage | MAP_RAW) & MAP_FLAGS);
-
-      for (int s = 0; s < box->depth; s++) {
-         unsigned x0_el, y0_el;
-         get_image_offset_el(surf, xfer->level, box->z + s, &x0_el, &y0_el);
-
-         for (uint32_t y = 0; y < box->height; y++) {
-            for (uint32_t x = 0; x < box->width; x++) {
-               ptrdiff_t offset = s8_offset(surf->row_pitch_B,
-                                            x0_el + box->x + x,
-                                            y0_el + box->y + y);
-               untiled_s8_map[s * xfer->layer_stride + y * xfer->stride + x] =
-                  tiled_s8_map[offset];
-            }
-         }
-      }
-   }
-
-   map->unmap = iris_unmap_s8;
-}
-
 /* Compute extent parameters for use with tiled_memcpy functions.
  * xs are in units of bytes and ys are in units of strides.
  */
@@ -2538,6 +2469,15 @@ iris_transfer_map(struct pipe_context *ctx,
       usage |= PIPE_MAP_UNSYNCHRONIZED;
    }
 
+   /* We are dealing with external memory object PIPE_BUFFER, disable
+    * async mapping because of sync issues.
+    */
+   if (!res->mod_info &&
+       res->external_format != PIPE_FORMAT_NONE &&
+       resource->target == PIPE_BUFFER) {
+      usage &= ~PIPE_MAP_UNSYNCHRONIZED;
+   }
+
    /* Avoid using GPU copies for persistent/coherent buffers, as the idea
     * there is to access them simultaneously on the CPU & GPU.  This also
     * avoids trying to use GPU copies for our u_upload_mgr buffers which
@@ -2624,10 +2564,7 @@ iris_transfer_map(struct pipe_context *ctx,
          }
       }
 
-      if (surf->tiling == ISL_TILING_W) {
-         /* TODO: Teach iris_map_tiled_memcpy about W-tiling... */
-         iris_map_s8(map);
-      } else if (surf->tiling != ISL_TILING_LINEAR) {
+      if (surf->tiling != ISL_TILING_LINEAR) {
          iris_map_tiled_memcpy(map);
       } else {
          iris_map_direct(map);
@@ -2696,7 +2633,7 @@ iris_transfer_unmap(struct pipe_context *ctx, struct pipe_transfer *xfer)
  * The pipe->texture_subdata() driver hook.
  *
  * Mesa's state tracker takes this path whenever possible, even with
- * PIPE_CAP_TEXTURE_TRANSFER_MODES set.
+ * pipe_caps.texture_transfer_modes set.
  */
 static void
 iris_texture_subdata(struct pipe_context *ctx,
@@ -2746,28 +2683,13 @@ iris_texture_subdata(struct pipe_context *ctx,
    for (int s = 0; s < box->depth; s++) {
       const uint8_t *src = data + s * layer_stride;
 
-      if (surf->tiling == ISL_TILING_W) {
-         unsigned x0_el, y0_el;
-         get_image_offset_el(surf, level, box->z + s, &x0_el, &y0_el);
+      unsigned x1, x2, y1, y2;
+      tile_extents(surf, box, level, s, &x1, &x2, &y1, &y2);
 
-         for (unsigned y = 0; y < box->height; y++) {
-            for (unsigned x = 0; x < box->width; x++) {
-               ptrdiff_t offset = s8_offset(surf->row_pitch_B,
-                                            x0_el + box->x + x,
-                                            y0_el + box->y + y);
-               dst[offset] = src[y * stride + x];
-            }
-         }
-      } else {
-         unsigned x1, x2, y1, y2;
-
-         tile_extents(surf, box, level, s, &x1, &x2, &y1, &y2);
-
-         isl_memcpy_linear_to_tiled(x1, x2, y1, y2,
-                                    (void *)dst, (void *)src,
-                                    surf->row_pitch_B, stride,
-                                    false, surf->tiling, ISL_MEMCPY);
-      }
+      isl_memcpy_linear_to_tiled(x1, x2, y1, y2,
+                                 (void *)dst, (void *)src,
+                                 surf->row_pitch_B, stride,
+                                 false, surf->tiling, ISL_MEMCPY);
    }
 }
 
@@ -2850,6 +2772,52 @@ static const struct u_transfer_vtbl transfer_vtbl = {
    .get_stencil           = iris_resource_get_separate_stencil,
 };
 
+static struct pipe_vm_allocation *
+iris_alloc_vm(struct pipe_screen *pscreen, uint64_t start, uint64_t size)
+{
+   struct iris_screen *screen = (struct iris_screen *)pscreen;
+   if (!iris_bufmgr_alloc_heap(screen->bufmgr, start, size))
+      return NULL;
+
+   struct pipe_vm_allocation *res = CALLOC_STRUCT(pipe_vm_allocation);
+   if (!res) {
+      iris_bufmgr_free_heap(screen->bufmgr, start, size);
+      return NULL;
+   }
+
+   res->start = start;
+   res->size = size;
+   return res;
+}
+
+static void
+iris_free_vm(struct pipe_screen *pscreen, struct pipe_vm_allocation *alloc)
+{
+   struct iris_screen *screen = (struct iris_screen *)pscreen;
+   iris_bufmgr_free_heap(screen->bufmgr, alloc->start, alloc->size);
+   FREE(alloc);
+}
+
+static bool
+iris_resource_assign_vma(struct pipe_screen *pscreen,
+                         struct pipe_resource *presource, uint64_t address)
+{
+   struct iris_screen *screen = (struct iris_screen *)pscreen;
+   struct iris_resource *res = (struct iris_resource *)presource;
+
+   assert(presource->flags & PIPE_RESOURCE_FLAG_FRONTEND_VM);
+   return iris_bufmgr_assign_vma(screen->bufmgr, res->bo, address);
+}
+
+static uint64_t
+iris_resource_get_address(struct pipe_screen *pscreen,
+                          struct pipe_resource *presrouce)
+{
+   struct iris_resource *res = (struct iris_resource *)presrouce;
+   assert(presrouce->flags & PIPE_RESOURCE_FLAG_FIXED_ADDRESS);
+   return res->bo->address + res->offset;
+}
+
 void
 iris_init_screen_resource_functions(struct pipe_screen *pscreen)
 {
@@ -2867,6 +2835,10 @@ iris_init_screen_resource_functions(struct pipe_screen *pscreen)
    pscreen->resource_destroy = u_transfer_helper_resource_destroy;
    pscreen->memobj_create_from_handle = iris_memobj_create_from_handle;
    pscreen->memobj_destroy = iris_memobj_destroy;
+   pscreen->alloc_vm = iris_alloc_vm;
+   pscreen->free_vm = iris_free_vm;
+   pscreen->resource_assign_vma = iris_resource_assign_vma;
+   pscreen->resource_get_address = iris_resource_get_address;
    pscreen->transfer_helper =
       u_transfer_helper_create(&transfer_vtbl,
                                U_TRANSFER_HELPER_SEPARATE_Z32S8 |
